/********************************************************************************************
 *
 * Copyright 2012-2018  Frank Gollnik  <http://www.jabotics.org>
 *
 * 
 * This file is part of the Java software package org.jabotics (in the following: Jabotics).
 * 
 * Jabotics is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Jabotics is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jabotics.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * Diese Datei ist Teil des Java-Software-Pakets org.jabotics (im Folgenden kurz: Jabotics).
 * 
 * Jabotics ist Freie Software: Sie können es unter den Bedingungen
 * der GNU General Public License, wie von der Free Software Foundation,
 * Version 3 der Lizenz oder (nach Ihrer Wahl) jeder späteren
 * veröffentlichten Version, weiterverbreiten und/oder modifizieren.
 * 
 * Jabotics wird in der Hoffnung, dass es nützlich sein wird, aber
 * OHNE JEDE GEWÄHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 * Gewährleistung der MARKTFÄHIGKEIT oder EIGNUNG FÜR EINEN BESTIMMTEN ZWECK.
 * Siehe die GNU General Public License für weitere Details.
 * 
 * Sie sollten eine Kopie der GNU General Public License zusammen mit Jabotics
 * erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
 * 
 *******************************************************************************************/

 
package org.jabotics.robot.core;

import org.jabotics.robot.compile.*;
import org.jabotics.robot.defs.*;
import org.jabotics.robot.util.*;
import org.jabotics.robot.adapter.XAdapter;

/**
 * Basisklasse für speziellere Sensorklassen wie XDrucksensor, XSchallsensor,
 * XLichtsensor und XUltraschallsensor. CSensor stellt allgemeine Basisfunktionalitäten
 * für die abgeleiteten Sensorklassen zur Verfügung. Die Standardimplementierung in
 * CSensor muss in den abgeleiteten Klassen ggf. passend überschrieben werden.
 * 
 * @author  Frank Gollnik
 * @version 1.1.1
 */
public class CSensor
{

    // constants and class variables
    static private final int UPDATE_PERIOD = 10;    // im ms
    static private final int UPDATE_MIN_DELAY = 5; // im ms
    static private final int CONTROLLER_THREAD_PRIORITY = Thread.MAX_PRIORITY - 3;
    static private final TVolatileRefArray<CSensor> sSensors = new TVolatileRefArray<CSensor>(D.TOTAL_SENSOR_NUMBER);
    static private CEventDispatcher sDispatcher = null;

    static private final IAdapterFactory sAdapterFactory;
    static private final CSensor sInvalidSensor;
    static private final ITimerAdapter sTimer;
    static private final XTimedThread sThread;
    static private long sSTRTime = 0;
    static private long sSTRLastTime = 0;

    // for session management (in conjunction with simulation module)
    static private volatile boolean sStopController = false;
    static private volatile boolean sControllerHalted = false;

    static
    {
        sAdapterFactory = XAdapter.getFactory();
        sInvalidSensor = new CSensor();
        sTimer = sAdapterFactory.getTimerAdapter();

        sThread = new XTimedThread("Sensor Controller Thread",new SensorThreadRunnable());
        sThread.setPriority(CONTROLLER_THREAD_PRIORITY);
        sThread.setDaemon(true);
        sThread.setTimeouts(UPDATE_PERIOD,UPDATE_MIN_DELAY,UPDATE_PERIOD); // Beispiel: Periode = 50ms entsprechen 1000ms/50ms = 20 Messungen pro Sekunde
        sThread.start();

        if (CS.SESSIONS_ENABLED)
        {
            ISessionAdapter session = sAdapterFactory.getSessionAdapter();
            session.add(new SessionDataManager());
        }
    }


    // instance variables
    private ISensorAdapter mSensorAdapter;
    private final float mPeriodSec;
    private long mTime;
    private int mUpdateCount = 0;
    private volatile boolean mInitialized = false;
    private volatile int mIntegrationTime = 0;
    private volatile float mIntegrationTimeSec = 0f;
    private volatile EventFilter mEventFilter = new EventFilter();
    private volatile EventFilter mNewEventFilter = new EventFilter();
    private float mMeasuredValue;   // only accessed in method update
    private float mIntegratedValue; // only accessed in method update
    private volatile float mOutputValue;
    private volatile float mChangeRate = 0f;
    private volatile float mLastReportedOutputValue;
    private volatile float mLastReportedChangeRate = 0f;
    private volatile CEventDispatcher.IHandler mHandler = null;
    private volatile CEventDispatcher.IHandler mNewHandler = null;


    static private class SessionDataManager implements ISessionData
    {
        public String getKey()
        {
            return "CSensor.SessionDataManager";
        }

        @Override
        public void stopSession()
        {
            try
            {
                sStopController = true;
                while (!sControllerHalted) Thread.sleep(1);
            }
            catch (Exception e) {}
        }

        @Override
        public ISessionData resetSession()
        {
            for (int i = 0;  i < D.TOTAL_SENSOR_NUMBER;  ++i)
            {
                CSensor sensor = sSensors.get(i);
                if ((null != sensor) && (sensor.mInitialized)) sensor.initialize(); // repeat initialization
            }
            return this;
        }

        @Override
        public void startSession()
        {
            sSTRTime = sTimer.currentTimeMillis();
            sSTRLastTime = sSTRTime;
            sStopController = false;
            sControllerHalted = false; // do not wait for controller thread to update sControllerHalted
        }
    }


    static private class EventFilter
    {
        volatile float mResolutionValue = 1f;
        volatile float mResolutionRate = 0f;
    }


    static private class SensorThreadRunnable implements XTimedThread.Runnable
    {
        @Override
        public void init()
        {
            sSTRTime = sTimer.currentTimeMillis();
            sSTRLastTime = sSTRTime;
        }

        @Override
        public void runTimed()
        {
            if (CS.SESSIONS_ENABLED)
            {
                sControllerHalted = sStopController;
                if (sStopController) return;
            }

            // Timer is pausing?
            sSTRTime = sTimer.currentTimeMillis();
            if (sSTRTime <= sSTRLastTime) return;
            sSTRLastTime = sSTRTime;
            
            for (int i = 0;  i < D.TOTAL_SENSOR_NUMBER;  ++i)
            {
                CSensor sensor = sSensors.get(i);
                if (null != sensor)  sensor.update();
            }

            for (int i = 0;  i < D.TOTAL_SENSOR_NUMBER;  ++i)
            {
                CSensor sensor = sSensors.get(i);
                if (null != sensor)  sensor.notifyHandlers();
            }

            if (CS.SENSOR_EVENT_DISPATCHER_ENABLED)
            {
                if (null != sDispatcher) sDispatcher.notifyOnSensorUpdates();
            }
        }

        @Override
        public void terminate()
        {
        }

        @Override
        public void shutdown()
        {
        }
    }

    static synchronized private CEventDispatcher getDispatcher()
    {
        if (CS.SENSOR_EVENT_DISPATCHER_ENABLED)
        {
            if (null == sDispatcher) sDispatcher = CEventDispatcher.getDispatcher();
            return sDispatcher;
        }
        else
        {
            return null;
        }
    }

    static synchronized public CSensor getSensor(ESensorPort port, ESensorType type)
    {
        int id = port.ordinal();
        CSensor sensor = sSensors.get(id);
        if (null == sensor)
        {
            ISensorAdapter adapter = sAdapterFactory.getSensorAdapter(port,type);
            if (ESensorType.INVALID_SENSOR == adapter.getType()) return sInvalidSensor;
            sensor = new CSensor(adapter);
            sSensors.set(id,sensor);
            return sensor;
        }

        if (sensor.getType() != type)
        {
            System.err.println("Error: Different sensor types assigned to the same sensor port.");
            return sInvalidSensor;
        }
        return sensor;
    }

    static synchronized public CSensor getButton(EButtonType type)
    {
        int id = D.SENSOR_PORT_NUMBER + type.ordinal();
        CSensor sensor = sSensors.get(id);
        if (null == sensor)
        {
            ISensorAdapter adapter = sAdapterFactory.getButtonAdapter(type);
            if (ESensorType.INVALID_SENSOR == adapter.getType()) return sInvalidSensor;
            sensor = new CSensor(adapter);
            sSensors.set(id,sensor);
        }
        return sensor; // has to be touch sensor
    }
    

    private CSensor()  // generates invalid sensor
    {
        mSensorAdapter = sAdapterFactory.getInvalidSensorAdapter();
        mPeriodSec = 0.001f * UPDATE_PERIOD * mSensorAdapter.getUpdatePeriodProlongationFactor();
    }

    private CSensor(ISensorAdapter adapter)
    {
        mSensorAdapter = adapter;
        mPeriodSec = 0.001f * UPDATE_PERIOD * mSensorAdapter.getUpdatePeriodProlongationFactor();
    }

    private void initialize()
    {
        mTime = sTimer.currentTimeMillis();
        mMeasuredValue = mSensorAdapter.getValue();
        mIntegratedValue = mMeasuredValue;
        mOutputValue = mMeasuredValue;
        mLastReportedOutputValue = mMeasuredValue;
        mUpdateCount = mSensorAdapter.getUpdatePeriodProlongationFactor();
        mInitialized = true;
    }

    private void update()
    {
        if (!mInitialized) initialize();

        if (--mUpdateCount > 0) return;
        mUpdateCount = mSensorAdapter.getUpdatePeriodProlongationFactor();

        float lastMeasuredValue = mMeasuredValue;
        mMeasuredValue = mSensorAdapter.getValue();

        long lastTime = mTime;
        mTime = sTimer.currentTimeMillis();
        float period = (mTime > lastTime) ? 0.001f * (float)(mTime - lastTime) : mPeriodSec;

        if (mIntegrationTime > 0)
        {
            float response = period / (mIntegrationTimeSec + period);
            float oneMinusResponse = 1f - response;
            float lastOutputValue = mOutputValue;
            mIntegratedValue = oneMinusResponse * mIntegratedValue + response * mMeasuredValue;
            if (isTouchSensorType())
            {
                mOutputValue = Math.round(mIntegratedValue);
                if (mOutputValue != lastOutputValue) mIntegratedValue = (float) mOutputValue;  // avoid fast jumping back to previous output value
                mChangeRate = oneMinusResponse * mChangeRate + response * (mOutputValue - lastOutputValue) / period;
            }
            else
            {
                mOutputValue = mIntegratedValue;
                mChangeRate = oneMinusResponse * mChangeRate + response * (mMeasuredValue - lastMeasuredValue) / period;
            }
        }
        else
        {
            mIntegratedValue = mMeasuredValue;
            mOutputValue = mMeasuredValue;
            mChangeRate = (mMeasuredValue - lastMeasuredValue) / period;
        }
    }

    private void notifyHandlers()
    {
        if (mHandler != mNewHandler)
        {
            if (null != mNewHandler)
            {
                mLastReportedOutputValue = mOutputValue;
                mLastReportedChangeRate = mChangeRate;
                // send initial event notification to new handler:
                if (CS.SENSOR_EVENT_DISPATCHER_ENABLED)
                {
                    sDispatcher.addEvent(mNewHandler, mSensorAdapter, mLastReportedOutputValue, mOutputValue, mLastReportedChangeRate, mChangeRate);
                }
                else
                {
                    mNewHandler.manageEvent(mSensorAdapter.getPortID(), mLastReportedOutputValue, mOutputValue, mLastReportedChangeRate, mChangeRate);
                }
            }
            mHandler = mNewHandler;
        }
        else if (null != mHandler)
        {
            if (changesExceedResolution())
            {
                if (CS.SENSOR_EVENT_DISPATCHER_ENABLED)
                {
                    sDispatcher.addEvent(mHandler, mSensorAdapter, mLastReportedOutputValue, mOutputValue, mLastReportedChangeRate, mChangeRate);
                }
                else
                {
                    mHandler.manageEvent(mSensorAdapter.getPortID(), mLastReportedOutputValue, mOutputValue, mLastReportedChangeRate, mChangeRate);
                }
                mLastReportedOutputValue = mOutputValue;
                mLastReportedChangeRate = mChangeRate;
            }
        }
    }
    
    private boolean changesExceedResolution()
    {
        // No filtering at all?
        if ( (0 == mEventFilter.mResolutionRate) && (0 == mEventFilter.mResolutionValue) ) return true;

        float valueChange = Math.abs(mOutputValue - mLastReportedOutputValue);
        // Filtering for value only?
        if (0 == mEventFilter.mResolutionRate) return (valueChange >= mEventFilter.mResolutionValue);

        float rateChange = Math.abs(mChangeRate - mLastReportedChangeRate);
        // Filtering for rate only?
        if (0 == mEventFilter.mResolutionValue) return (rateChange >= mEventFilter.mResolutionRate);

        // Filtering for value and rate: Send notification when at least one relevant change
        return ( (valueChange >= mEventFilter.mResolutionValue) || (rateChange >= mEventFilter.mResolutionRate) );
    }

    public ESensorType getType()
    {
        return mSensorAdapter.getType();
    }

    public boolean isTouchSensorType()
    {
        return (mSensorAdapter.getType() == ESensorType.TOUCH_SENSOR);
    }

    public boolean hasLight()
    {
        return mSensorAdapter.hasLight();
    }

    public void lightOn()
    {
        mSensorAdapter.lightOn();
    }

    public void lightOff()
    {
        mSensorAdapter.lightOff();
    }

    public float getOutputValue()
    {
        while (!mInitialized) sTimer.msDelay(2);
        return mOutputValue;
    }

    public float getChangeRate()
    {
        return mChangeRate;
    }
    
    public int getIntegrationTime()
    {
        return mIntegrationTime;
    }
    
    public float getEventFilterValue()
    {
        return mEventFilter.mResolutionValue;
    }
    
    public float getEventFilterRate()
    {
        return mEventFilter.mResolutionRate;
    }

    public void setIntegrationTime(int integration_time)
    {
        if (integration_time > 0)
        {
            mIntegrationTime = integration_time;
            mIntegrationTimeSec = 0.001f * (float)integration_time;
        }
        else 
        {
            mIntegrationTime = 0;
            mIntegrationTimeSec = 0f;
        }
    }

    public void resetEventFilter()
    {
        setEventFilter(1f,0f);
    }

    public synchronized void setEventFilter(float resolutionValue, float resolutionRate)
    {
        mNewEventFilter.mResolutionValue = (resolutionValue > 0) ? resolutionValue : 0;
        mNewEventFilter.mResolutionRate = (resolutionRate > 0) ? resolutionRate : 0;
        // toggle filter instances:
        EventFilter filter = mEventFilter;
        mEventFilter = mNewEventFilter;
        mNewEventFilter = filter;
    }

    public synchronized void setEventHandler(CEventDispatcher.IHandler handler)
    {
        if (handler != mHandler)
        {
            if (CS.SENSOR_EVENT_DISPATCHER_ENABLED)
            {
                getDispatcher(); // initializes the dispatcher if not yet done
            }

            mNewHandler = handler;
            // Make sure that no more event notifications will be sent to the previous handler:
            if (null != mHandler)
            {
                while (mNewHandler != mHandler) sTimer.msDelay(1);
                // The sensor will send no more event notifications for the previous handler to the dispatcher.
            }

            if (CS.SENSOR_EVENT_DISPATCHER_ENABLED)
            {
                sDispatcher.setEventHandler(mSensorAdapter, handler);
                // The dispatcher now has sent all remaining event notifications to the previous handler
                // and at least one initial notification to the new handler (if not null).
            }
        }
    }

    public boolean getAndResetOverflowFlag()
    {
        if (CS.SENSOR_EVENT_DISPATCHER_ENABLED)
        {
            return (null == sDispatcher) ? false : sDispatcher.getAndResetOverflowFlag(mSensorAdapter);
        }
        else
        {
            return false;
        }
    }

    public void waitUntilBeyondLimit(float limit)
    {
        while (limit >= mOutputValue)
        {
            sTimer.msDelay(2);
        }
    }

    public void waitUntilBelowLimit(float limit)
    {
        while (limit <= mOutputValue)
        {
            sTimer.msDelay(2);
        }
    }

    public boolean waitUntilBeyondLimit(float limit, int timeout)
    {
        long end = sTimer.currentTimeMillis() + timeout;
        while (limit >= mOutputValue)
        {
            if (sTimer.currentTimeMillis() >= end) return false;
            sTimer.msDelay(2);
        }
        return true;
    }

    public boolean waitUntilBelowLimit(float limit, int timeout)
    {
        long end = sTimer.currentTimeMillis() + timeout;
        while (limit <= mOutputValue)
        {
            if (sTimer.currentTimeMillis() >= end) return false;
            sTimer.msDelay(2);
        }
        return true;
    }

    public boolean isPressed()
    {
        return (0f != getOutputValue());
    }

    public void waitUntilPressed()
    {
        waitUntilBeyondLimit(0.5f);
    }

    public void waitUntilReleased()
    {
        waitUntilBelowLimit(0.5f);
    }

    public boolean waitUntilPressed(int timeout)
    {
        return waitUntilBeyondLimit(0.5f,timeout);
    }

    public boolean waitUntilReleased(int timeout)
    {
        return waitUntilBelowLimit(0.5f,timeout);
    }

    public void waitForPressEvent()
    {
        waitUntilBelowLimit(0.5f);
        waitUntilBeyondLimit(0.5f);
    }

    public void waitForReleaseEvent()
    {
        waitUntilBeyondLimit(0.5f);
        waitUntilBelowLimit(0.5f);
    }

    public boolean waitForPressEvent(int timeout)
    {
        long start = sTimer.currentTimeMillis();
        boolean released = waitUntilBelowLimit(0.5f,timeout);
        if (!released) return false;
        int remaining = timeout - (int) (sTimer.currentTimeMillis() - start);
        return (remaining > 0) ? waitUntilBeyondLimit(0.5f,remaining) : false;
    }

    public boolean waitForReleaseEvent(int timeout)
    {
        long start = sTimer.currentTimeMillis();
        boolean pressed = waitUntilBeyondLimit(0.5f,timeout);
        if (!pressed) return false;
        int remaining = timeout - (int) (sTimer.currentTimeMillis() - start);
        return (remaining > 0) ? waitUntilBelowLimit(0.5f,remaining) : false;
    }
}