/********************************************************************************************
 *
 * Copyright 2012-2018  Frank Gollnik  <http://www.jabotics.org>
 *
 * 
 * This file is part of the Java software package org.jabotics (in the following: Jabotics).
 * 
 * Jabotics is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Jabotics is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jabotics.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * Diese Datei ist Teil des Java-Software-Pakets org.jabotics (im Folgenden kurz: Jabotics).
 * 
 * Jabotics ist Freie Software: Sie können es unter den Bedingungen
 * der GNU General Public License, wie von der Free Software Foundation,
 * Version 3 der Lizenz oder (nach Ihrer Wahl) jeder späteren
 * veröffentlichten Version, weiterverbreiten und/oder modifizieren.
 * 
 * Jabotics wird in der Hoffnung, dass es nützlich sein wird, aber
 * OHNE JEDE GEWÄHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 * Gewährleistung der MARKTFÄHIGKEIT oder EIGNUNG FÜR EINEN BESTIMMTEN ZWECK.
 * Siehe die GNU General Public License für weitere Details.
 * 
 * Sie sollten eine Kopie der GNU General Public License zusammen mit Jabotics
 * erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
 * 
 *******************************************************************************************/


package org.jabotics.robot.en;


/**
 * <font color="LimeGreen"><b>&bull;</b></font> The interface ISensor provides methods for the configuration and readout of sensors.
 * 
 * @author  Frank Gollnik
 * @version 1.1.1
 */

public interface ISensor
{
    /**
     * <font color="LimeGreen"><b>&bull;</b></font> Returns a single measurand of the sensor. 
     * The possible values depend on the type of sensor:
     * value 0 (not pressed) or 1 (pressed) for touch sensors,
     * range of values from 0 (silent) to 100 (loud) for sound sensors,
     * range of values from 0 (dark) to 100 (bright) for light sensors,
     * distance to the next barrier in cm for ultrasonic sensors.
     * Note that the maximum range of the ultrasonic sensor is
     * typically limited to about 250 cm.
     * 
     * @return      measurand of the sensor
     */
    public float getMeasuredValue();

    /**
     * <font color="Gold"><b>&bull;</b></font> Returns the change rate for the measurand.
     * Sensors are steadily monitoring changes of their measurand in order to calculate
     * the corresponding change rate.
     * 
     * @return  change rate of measured values (in units per second)
     */
    public float getChangeRate();

    /**
     * <font color="Red"><b>&bull;</b></font> Returns the currently stored tolerance limit
     * for the sensor event filtering by measurand.
     * Changes of the measurand reaching the tolerance limit will be reported together with
     * the current change rate to a registered sensor listener by invoking
     * {@link ISensorListener#manageEvent(ISensor,float,float,float,float)}.
     * Since the tolerance limit must not exceed the possible range of measured values, it is fixed
     * to a value of 1 for touch sensors. For other sensors, the tolerance limit is set to 1
     * by default, since this is in general the resolution of a sensor's raw data.
     * A tolerance value of 0 will turn off the filtering by measurand. Sensor data will be passed
     * to a listener at the sensor's measuring rate in this case unless the filtering by change rate
     * has been turned on.
     * 
     * @return      the currently stored tolerance limit for the event filtering by measurand
     * @see         #getFilterToleranceRate()
     * @see         #setFilterTolerances(float,float)
     * @see         ISensorListener
     */
    public float getFilterToleranceValue();
    

    /**
     * <font color="Red"><b>&bull;</b></font> Returns the currently stored tolerance limit
     * for the sensor event filtering by change rate.
     * Changes of the change rate reaching the tolerance limit will be reported together with
     * the current measurand to a registered sensor listener by invoking
     * {@link ISensorListener#manageEvent(ISensor,float,float,float,float)}.
     * The default value for this tolerance is 0, which means that no filtering by change rate
     * will be done for sensor events. In addition to filtering by change rate, filtering by
     * measurand has to be taken into account.
     * 
     * @return      the currently stored tolerance limit for the event filtering by change rate
     * @see         #getFilterToleranceValue()
     * @see         #setFilterTolerances(float,float)
     * @see         ISensorListener
     */
    public float getFilterToleranceRate();

    /**

     * <font color="Gold"><b>&bull;</b></font> Returns the integration time constant which determines the inertia of the sensor.
     * The measurand of the sensor returned by {@link #getMeasuredValue()} represents a mean value
     * that will be calculated over the integration time (in ms) from the sensor's raw data.
     * If the integration time is large, the response of the sensor will be slow.
     * The integration time will also influence the calculated change rate (see {@link #getChangeRate()}).
     * The default value for the integration time is 0 (ms), that is, integration is turned off
     * corresponding to the fastest possible response of the sensor.
     * 
     * @return        the currently stored integration time of the sensor (in ms)
     */
    public int getIntegrationTime();
    
    /**
     * <font color="Red"><b>&bull;</b></font> Sets the tolerance limits for indicating changes
     * of the measurand and its change rate.
     * For a sensor event to be reported to a listener, at least one of the deviations
     * (measurand or change rate) has to reach or exceed the corresponding tolerance limit.
     * A tolerance of 0 will completely turn off the corresponding event filtering.
     * 
     * @param  tolerance_value   the new tolerance limit for changes of the measurand
     * @param  tolerance_rate    the new tolerance limit for changes of the change rate
     * @see    #getFilterToleranceValue()
     * @see    #getFilterToleranceRate()
     */
    public void setFilterTolerances(float tolerance_value, float tolerance_rate);

    /**
     * <font color="Gold"><b>&bull;</b></font> Sets the integration time of the sensor.
     * 
     * @param  time    the new integration time of the sensor (in ms)
     * @see    #getIntegrationTime()
     */
    public void setIntegrationTime(int time);

    /**
     * <font color="LimeGreen"><b>&bull;</b></font> Waits until the measurand exceeds the given value.
     * 
     * @param  limit    the limit that needs to be exceeded by the measurand for the method to return to the caller
     */
    public void waitUntilBeyondLimit(float limit);

    /**
     * <font color="LimeGreen"><b>&bull;</b></font> Waits until the measurand has fallen below the given value.
     * 
     * @param  limit    the limit that the measurand has to fall below for the method to return to the caller
     */
    public void waitUntilBelowLimit(float limit);

    /**
     * <font color="LimeGreen"><b>&bull;</b></font> Waits until the measurand exceeds the given value or the timeout period has passed.
     * 
     * @param  limit    the limit that should be exceeded by the measurand for the method to return to the caller
     * @param  timeout  the maximum time period to wait for the measurand to exceed the given limit
     * 
     * @return  true, if the limit has been exceeded or false, if not
     */
    public boolean waitUntilBeyondLimit(float limit, int timeout);

    /**
     * <font color="LimeGreen"><b>&bull;</b></font> Waits until the measurand has fallen below the given value or the timeout period has passed.
     * 
     * @param  limit    the limit that the measurand should fall below for the method to return to the caller
     * @param  timeout  the maximum time period to wait for the measurand to fall below the given limit
     * 
     * @return  true, if the measurand has fallen below the limit or false, if not
     */
    public boolean waitUntilBelowLimit(float limit, int timeout);

    /**
     * <font color="Red"><b>&bull;</b></font> Registers a sensor listener.
     * A previously registered sensor listener will be replaced by the new one.
     * During the method call, all possibly remaining event notifications for
     * the previous sensor listener will be processed, as well as an initial
     * event notification for the new sensor listener. The execution of a call
     * to this method will take some time (typically 10 ms or more) due to the
     * necessary synchronization of involved background processes.
     * 
     * @param  listener   the sensor listener that will receive event notifications on changes of the measurand
     * 
     * @see    #clearListener()
     */
    public void setListener(ISensorListener listener);
    
    /**
     * <font color="Red"><b>&bull;</b></font> Clears a registered sensor listener.
     * Similar conditions apply as for {@link #setListener(ISensorListener)}.
     * 
     * @see    #setListener(ISensorListener)
     */
    public void clearListener();
    
    /**
     * <font color="Red"><b>&bull;</b></font> Checks for an overflow of the event handling and resets the overflow flag.
     * This method can be called by an event listener to check whether some event notifications had to be dropped due
     * to an overflow of the event queue.
     * 
     * @return  true, if an overflow of the event handling occurred or false, if not
     * @see     ISensorListener
     */
    public boolean checkForEventOverflow();
}
