/********************************************************************************************
 *
 * Copyright 2012-2018  Frank Gollnik  <http://www.jabotics.org>
 *
 * 
 * This file is part of the Java software package org.jabotics (in the following: Jabotics).
 * 
 * Jabotics is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Jabotics is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jabotics.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * Diese Datei ist Teil des Java-Software-Pakets org.jabotics (im Folgenden kurz: Jabotics).
 * 
 * Jabotics ist Freie Software: Sie können es unter den Bedingungen
 * der GNU General Public License, wie von der Free Software Foundation,
 * Version 3 der Lizenz oder (nach Ihrer Wahl) jeder späteren
 * veröffentlichten Version, weiterverbreiten und/oder modifizieren.
 * 
 * Jabotics wird in der Hoffnung, dass es nützlich sein wird, aber
 * OHNE JEDE GEWÄHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 * Gewährleistung der MARKTFÄHIGKEIT oder EIGNUNG FÜR EINEN BESTIMMTEN ZWECK.
 * Siehe die GNU General Public License für weitere Details.
 * 
 * Sie sollten eine Kopie der GNU General Public License zusammen mit Jabotics
 * erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
 * 
 *******************************************************************************************/


package org.jabotics.robot.util;


/**
 * Use template class TGate to pass updates of a data object of type T from a writing thread to a reading thread.
 * Internally three data object instances are being used to buffer different object states. When writing data
 * to the gate object, the data of the passed template object will be copied to one of the buffer objects.
 * When reading from the gate object, the buffer object with the latest update will be returned. This buffer
 * instance will remain unchanged until a new read operation will occure on the gate object. The write method
 * returns a reference to the written buffer element which in general should not be modified afterwards.
 * In order to avoid copy operations, a reference to a free buffer object can be obtained by invocation of
 * getWriteObject. After directly manipulating this object, it can be committed with commitWriteObject.
 * No other write operations to the TGate instance must occur in the meantime. 
 * 
 * @author  Frank Gollnik
 * @version 1.1.1
 */
public class TGate<T extends TICloneable<T>>
{
    private class Link<T>
    {
        private Link<T> mNext = null;
        final private T mData;
        
        private Link(T data)
        {
            mData = data;
        }
    }

    private volatile Link<T> mRead;
    private volatile Link<T> mWrite;
    private volatile Link<T> mWorking;

    public TGate(T template)
    {
        mRead = new Link<T>(template.getClone());
        mWrite = new Link<T>(template.getClone());
        mWorking = new Link<T>(template.getClone());
        mRead.mNext = mWrite;
        mWrite.mNext = mWorking;
        mWorking.mNext = mRead;
        mRead = mWrite;
    }

    public boolean isModified()
    {
        return (mRead != mWrite);
    }

    public T read()
    {
        while (mRead != mWrite) mRead = mWrite;
        return mRead.mData;
    }

    public T write(T template)
    {
        while ((mWorking == mWrite) || (mWorking == mRead)) mWorking = mWorking.mNext; 
        mWorking.mData.copyDataFrom(template);
        return (mWrite = mWorking).mData;
    }

    public T getWriteObject()
    {
        while ((mWorking == mWrite) || (mWorking == mRead)) mWorking = mWorking.mNext; 
        return mWorking.mData;
    }

    public T commitWriteObject()
    {
        return ((mWorking == mWrite) || (mWorking == mRead)) ? null : (mWrite = mWorking).mData;
    }
}
