/********************************************************************************************
 *
 * Copyright 2012-2018  Frank Gollnik  <http://www.jabotics.org>
 *
 * 
 * This file is part of the Java software package org.jabotics (in the following: Jabotics).
 * 
 * Jabotics is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Jabotics is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jabotics.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * Diese Datei ist Teil des Java-Software-Pakets org.jabotics (im Folgenden kurz: Jabotics).
 * 
 * Jabotics ist Freie Software: Sie können es unter den Bedingungen
 * der GNU General Public License, wie von der Free Software Foundation,
 * Version 3 der Lizenz oder (nach Ihrer Wahl) jeder späteren
 * veröffentlichten Version, weiterverbreiten und/oder modifizieren.
 * 
 * Jabotics wird in der Hoffnung, dass es nützlich sein wird, aber
 * OHNE JEDE GEWÄHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 * Gewährleistung der MARKTFÄHIGKEIT oder EIGNUNG FÜR EINEN BESTIMMTEN ZWECK.
 * Siehe die GNU General Public License für weitere Details.
 * 
 * Sie sollten eine Kopie der GNU General Public License zusammen mit Jabotics
 * erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
 * 
 *******************************************************************************************/


package org.jabotics.robot.util;


/**
 * This class provides a mechanism for passing data objects written by one thread
 * to another reading thread and at the same time dealing with the missing garbage
 * collection on NXT robots by means of object recycling. The method getSize returns
 * the number of data elements that can be read from the pipe. The method getCapacity
 * returns the number of elements in the pipe's data buffer. Read operations will
 * return on object refence to the next element in the pipe. The last element
 * read from the pipe will be kept constant until the next read operation occurs.
 * Afterwards it will be subject to object recycling. It is in the responsibility
 * of the user to make his own copies of read objects if necessary.
 * The constructor for a pipe object requires an initial buffer element of type T.
 * The write method returns a reference to the written buffer element which in
 * general should not be modified afterwards.
 * In order to avoid copy operations, a reference to a free buffer element can be
 * obtained by invocation of getWriteObject. After directly manipulating this object,
 * it can be finally added to the pipe with commitWriteObject.
 * No other write operations to the pipe instance must occur in the meantime. 
 *
 *
 * // typical preparations for usage:
 * 
 *  class MyDataClass implements TPipe.IData<MyDataClass>
 *  {
 *      float mValue1 = ...;
 *      float mValue2 = ...;
 *      ...
 *
 *      public MyDataClass()  // if any constructor
 *      {
 *          ...
 *      }
 *
 *      @Override
 *      public void copyDataFrom(MyDataClass template)
 *      {
 *          if (this != template)
 *          {
 *              mValue1 = template.mValue1;
 *              mValue2 = template.mValue2;
 *              ...
 *          }
 *      }
 *
 *      @Override
 *      public MyDataClass getCopy()
 *      {
 *          MyDataClass data = new MyDataClass();
 *          data.copyDataFrom(this);
 *          return data;
 *      }
 *  }
 * 
 * @author  Frank Gollnik
 * @version 1.1.1
 */
public class TPipe<T extends TICloneable<T>>
{
    private volatile int mCapacity = 1;
    private volatile Link<T> mRead;
    private volatile Link<T> mWrite;

    private class Link<T>
    {
        private Link<T> mNext = this;
        private int     mPosition = 0;
        private final T mData;

        private Link(T data)
        {
            mData = data;
        }
    }

    public TPipe(T template)
    {
        mWrite = new Link<T>(template.getClone());
        mRead = mWrite;
    }

    public int getCapacity()
    {
        return mCapacity;
    }

    public int getSize()
    {
        return mWrite.mPosition - mRead.mPosition;
    }

    public int getPosition()
    {
        return mRead.mPosition;
    }

    public boolean hasNext()
    {
        return (mRead != mWrite);
    }

    public T read()
    {
        if (mRead == mWrite) return null;
        mRead = mRead.mNext;
        return mRead.mData;
    }

    public T write(T template)
    {
        Link<T> next = getNextWriteElement(template);
        if (next == mWrite.mNext)  next.mData.copyDataFrom(template);
        return (mWrite = next).mData;
    }

    public T getWriteObject()
    {
        return getNextWriteElement(mWrite.mData).mData;
    }

    public T commitWriteObject()
    {
        Link<T> next = mWrite.mNext;
        return (next == mRead) ? null : (mWrite = next).mData;
    }

    private Link<T> getNextWriteElement(T template)
    {
        Link<T> next = mWrite.mNext;
        if (next == mRead)
        {
            next = new Link<T>(template.getClone());
            next.mNext = mRead;
            mWrite.mNext = next;
            ++mCapacity;
        }
        next.mPosition = mWrite.mPosition + 1;
        return next;
    }
}
