/********************************************************************************************
 *
 * Copyright 2012-2018  Frank Gollnik  <http://www.jabotics.org>
 *
 * 
 * This file is part of the Java software package org.jabotics (in the following: Jabotics).
 * 
 * Jabotics is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Jabotics is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jabotics.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * Diese Datei ist Teil des Java-Software-Pakets org.jabotics (im Folgenden kurz: Jabotics).
 * 
 * Jabotics ist Freie Software: Sie können es unter den Bedingungen
 * der GNU General Public License, wie von der Free Software Foundation,
 * Version 3 der Lizenz oder (nach Ihrer Wahl) jeder späteren
 * veröffentlichten Version, weiterverbreiten und/oder modifizieren.
 * 
 * Jabotics wird in der Hoffnung, dass es nützlich sein wird, aber
 * OHNE JEDE GEWÄHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 * Gewährleistung der MARKTFÄHIGKEIT oder EIGNUNG FÜR EINEN BESTIMMTEN ZWECK.
 * Siehe die GNU General Public License für weitere Details.
 * 
 * Sie sollten eine Kopie der GNU General Public License zusammen mit Jabotics
 * erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
 * 
 *******************************************************************************************/


package org.jabotics.robot.util;

import org.jabotics.robot.compile.*;
import java.util.LinkedList;
import java.util.ArrayList;
import java.util.Iterator;

/**
 * Die Klasse XTimedThread bietet einen Mechanismus zum regelmäßigen zeitgesteuerten Aufruf einer
 * runTimed()-Methode der an den Konstruktor von XTimedThread übergebenen XTimedThread.Runnable-Instanz.
 * Mit einem setPeriods(long period,long min_period)-Aufruf kann das Zeitintervall vorgegeben werden,
 * in dem die runTimed()-Methode aufgerufen wird, wobei min_period das minimale Zeitintervall ist,
 * das nach Durchlaufen der runTimed-Methode vor einem erneuten Aufruf abgewartet wird.
 * Vor dem ersten Aufruf von runTimed() wird das Runnable-Objekt durch Aufruf der init()-Methode initialisiert.
 * Dieser Aufruf erfolgt aus dem Timer-Thread heraus, so dass der Thread mit Thread.currentThread()
 * z.B. zur späteren Verwendung in terminate() (Aufruf aus einem anderen Thread!) ermittelt werden kann.
 * Die runTimed()-Methode kann auch gezielt durch einen Aufruf der Methode trigger() angestoßen werden.
 * Durch Aufruf der Methode terminate() wird der Thread geordnet gestoppt. Dabei wird die Runnable-Instanz
 * durch Aufruf der terminate()-Methode davon in Kenntnis gesetzt, dass der Thread gestoppt werden soll.
 * Die Runnable-Instanz kann so bei Bedarf eine beschleunigte Beendigung eines runTimed()-Durchlaufs veranlassen.
 * Es wird von XTimedThread zusätzlich ein Interrupt auf den Thread ausgelöst für den Fall, dass sich
 * der Thread in einem Wartezustand befindet. Nach Beendigung der Timer-Schleife wird die shutdown()-Methode
 * der Runnable-Instanz aufgerufen, so dass hier notwendige Aufräumarbeiten erledigt werden können.
 * Spätestens bei Programmende werden diese Terminierungsschritte druchlaufen.
 * 
 * Implemention has to take into account limitations of class java.lang.Thread for NXT.
 * 
 * @author  Frank Gollnik
 * @version 1.1.1
 */
public class XTimedThread
{
    public interface Runnable
    {
        void init();
        void runTimed();
        void terminate();
        void shutdown();
    }

    private class XRun implements java.lang.Runnable
    {
        public void run()
        {
            XTimedThread.this.run();
        }
    }

    private class XShutdownHook extends Thread
    {
        @Override
        public void run()
        {
            if (mThread.isAlive())
            {
                terminate();
                try
                {
                    mThread.join();
                }
                catch (InterruptedException e) {}
            }
        }
    }

    private static class Dependency
    {
        private volatile XTimedThread mThread;
        private final String          mThreadID;
        private volatile int          mMinCalls;
        private long                  mTargetCalls;

        private Dependency(String thread_id, int min_calls)
        {
            mThread = null;
            mThreadID = thread_id;
            mMinCalls = (min_calls < 1) ? 1 : min_calls;
            mTargetCalls = mMinCalls;
        }

        private Dependency(XTimedThread thread, int min_calls)
        {
            mThread = thread;
            mThreadID = thread.mID;
            mMinCalls = (min_calls < 1) ? 1 : min_calls;
            mTargetCalls = mMinCalls;
        }
    }

    // constants / class variables:
    private static final long  MILLION = 1000000;
    //private static final float MILLION_F = MILLION;
    private static final long  sStartNanos = System.nanoTime();
    // class variables for load balancing mechanism:
    private static final ArrayList<XTimedThread> sRegisteredThreads;
    private static final ArrayList<Dependency>   sPendingDependencies;
    private static volatile float sTimeDilatation = 1f;
    static
    {
        if (CS.TIMED_THREAD_DEPENDENCIES)
        {
            sRegisteredThreads = new ArrayList<XTimedThread>();
            sPendingDependencies = new ArrayList<Dependency>();
        }
        else
        {
            sRegisteredThreads = null;
            sPendingDependencies = null;
        }
    }
    
    // Die Instanzvariablen:
    private final Thread mThread = new Thread(new XRun());
    private final Object mSynchObject = new Object();
    private final Object mStartMonitor = new Object();
    private final Object mTerminateMonitor = new Object();
    private final XShutdownHook mShutdownHook = new XShutdownHook();
    private final LinkedList<Dependency> mDependencies = (CS.TIMED_THREAD_DEPENDENCIES ? new LinkedList<Dependency>() : null);
    private final String mID;
    private volatile Runnable mRunnable = null;
    private volatile boolean mTimeDilatationEnabled = CS.TIMED_THREAD_DILATATION;
    private volatile boolean mStarted = false;
    private volatile boolean mTerminate = false;
    private volatile boolean mTrigger = false;
    private volatile long    mCalls = 0;
    private volatile long    mTargetTime = 0;
    private volatile int     mPeriodMillis = 10;
    private volatile int     mMinDelayMillis = 5;
    private volatile int     mInitialDelayMillis = 0;
    private volatile long    mPeriodL;
    private volatile long    mMinDelayL;
    private volatile long    mInitialDelayL;
    private volatile float   mPeriodF;
    private volatile float   mMinDelayF;
    private volatile float   mInitialDelayF;

    public XTimedThread(Runnable runnable)
    {
        if (null == runnable) throw new IllegalArgumentException("XTimedThread: Runnable is null");
        mRunnable = runnable;
        mID = null;
        initTimeouts();
    }

    public XTimedThread(String id, Runnable runnable)
    {
        if (null == runnable) throw new IllegalArgumentException("XTimedThread: Runnable is null");
        mRunnable = runnable;
        mID = id;
        if (null != mID)
        {
            mThread.setName(mID);
            registerThread(this);
        }
        initTimeouts();
    }

    private void initTimeouts()
    {
        if (CS.TIMED_THREAD_NANOSECONDS)
        {
            mPeriodL = MILLION * (long)mPeriodMillis;
            mMinDelayL = MILLION * (long)mMinDelayMillis;
            mInitialDelayL = MILLION * (long)mInitialDelayMillis;
        }
        else
        {
            mPeriodL = (long)mPeriodMillis;
            mMinDelayL = (long)mMinDelayMillis;
            mInitialDelayL = (long)mInitialDelayMillis;
        }

        mPeriodF = (float)mPeriodL;
        mMinDelayF = (float)mMinDelayL;
        mInitialDelayF = (float)mInitialDelayL;
    }

    public void start()
    {
        if (!mStarted)
        {
            synchronized(mStartMonitor)
            {
                if (!mStarted)
                {
                    Runtime.getRuntime().addShutdownHook(mShutdownHook);
                    mThread.start();
                    mStarted = true;
                }
            }
        }
    }

    public String getId()
    {
        return mID;
    }

    public String getName()
    {
        return mThread.getName();
    }

    public void setName(String name)
    {
        mThread.setName(name);
    }

    public int getPriority()
    {
        return mThread.getPriority();
    }

    public void setPriority(int priority)
    {
        mThread.setPriority(priority);
    }

    public void setDaemon(boolean on)
    {
        mThread.setDaemon(on);
    }

    public boolean timeDilatationEnabled()
    {
        return mTimeDilatationEnabled;
    }

    public void enableTimeDilatation(boolean enabled)
    {
        mTimeDilatationEnabled = CS.TIMED_THREAD_DILATATION && enabled;
    }

    public void terminate()
    {
        if (mStarted && !mTerminate)
        {
            synchronized(mTerminateMonitor)
            {
                if (!mTerminate)
                {
                    mRunnable.terminate();
                    mTerminate = true;
                    mThread.interrupt();
                }
            }
        }
    }

    public void trigger()
    {
        synchronized(mSynchObject)
        {
            mTrigger = true;
            mSynchObject.notify();
        }
    }

    public int getPeriod()
    {
        return mPeriodMillis;
    }

    public int getMinDelay()
    {
        return mMinDelayMillis;
    }

    public int getInitialDelay()
    {
        return mInitialDelayMillis;
    }

    public void setTimeouts(int period, int min_delay, int initial_delay)
    {
        synchronized(mSynchObject)
        {
            mPeriodMillis = (period > 0) ? period : 1;
            mMinDelayMillis = (min_delay < 0) ? 0 : min_delay;
            mInitialDelayMillis = (initial_delay < 0) ? 0 : initial_delay;

            initTimeouts();
        }
    }

    public void addDependency(String thread_id, int min_calls)
    {
        if (CS.TIMED_THREAD_DEPENDENCIES)
        {
            synchronized(mDependencies)
            {
                Iterator<Dependency> it_dep = mDependencies.iterator();
                while (it_dep.hasNext())
                {
                    Dependency dep = it_dep.next();
                    if (dep.mThreadID.equals(thread_id))
                    {
                        dep.mMinCalls = (min_calls < 1) ? 1 : min_calls;
                        return;
                    }
                }
                mDependencies.add(getDependency(thread_id, min_calls));
            }
        }
    }

    public void addDependency(XTimedThread thread, int min_calls)
    {
        if (CS.TIMED_THREAD_DEPENDENCIES)
        {
            synchronized(mDependencies)
            {
                Iterator<Dependency> it_dep = mDependencies.iterator();
                while (it_dep.hasNext())
                {
                    Dependency dep = it_dep.next();
                    if (dep.mThread.equals(thread))
                    {
                        dep.mMinCalls = (min_calls < 1) ? 1 : min_calls;
                        return;
                    }
                }
                mDependencies.add(new Dependency(thread, min_calls));
            }
        }
    }

    private void run()
    {
        try
        {
            mRunnable.init();

            if (CS.TIMED_THREAD_NANOSECONDS)
            {
                mTargetTime = nanoTime();
            }
            else
            {
                mTargetTime = System.currentTimeMillis();
            }

            if (mInitialDelayL > 0)
            {
                mTargetTime += ((CS.TIMED_THREAD_DILATATION && mTimeDilatationEnabled) ? (long)(sTimeDilatation * mInitialDelayF) : mInitialDelayL);
                waitUntilTargetTime();
            }
            callRunTimed(); // initial call
            
            while(!mTerminate && !mThread.isInterrupted())
            {
                synchronized(mSynchObject)
                {
                    if (!mTrigger)  waitUntilTargetTime();
                }
                callRunTimed();
            }
        }
        catch(InterruptedException e)
        {
            if (CS.LOGGING_UTIL)
            {
                if (XLogBuffer.isEnabled())
                {
                    synchronized(XLogBuffer.getMonitor())
                    {
                        XLogBuffer.print("InterruptedException in XTimedThread: ");
                        XLogBuffer.println(this.mThread.getName());
                        XLogBuffer.print("Message: ");
                        XLogBuffer.println(e.getMessage());
                    }
                }
            }
        }

        mRunnable.shutdown();

        if (null != mID) unregisterThread(this);
    }

    private void waitUntilTargetTime() throws InterruptedException
    {
        if (CS.TIMED_THREAD_NANOSECONDS)
        {
            long currentNanos = nanoTime();
            if (mTargetTime > currentNanos)
            {
                long timeout = mTargetTime - currentNanos;
                int nanos = (int)(timeout % MILLION);
                long millis = (timeout - nanos) / MILLION;
                synchronized(mSynchObject)
                {
                    mSynchObject.wait(millis,nanos);
                }
            }
        }
        else
        {
            long currentMillis = System.currentTimeMillis();
            if (mTargetTime > currentMillis)
            {
                synchronized(mSynchObject)
                {
                    mSynchObject.wait(mTargetTime - currentMillis);
                }
            }
        }
    }

    private void callRunTimed() throws InterruptedException
    {
        if (CS.TIMED_THREAD_DEPENDENCIES)
        {
            Iterator<Dependency> it_dep = mDependencies.iterator();
            while (it_dep.hasNext())
            {
                Dependency dep = it_dep.next();
                while (dep.mTargetCalls > dep.mThread.mCalls)
                {
                    synchronized(mSynchObject)
                    {
                        mSynchObject.wait(1);
                    }
                }
                dep.mTargetCalls = dep.mThread.mCalls + dep.mMinCalls;
            }
        }

        if (CS.LOGGING_UTIL)
        {
            if (XLogBuffer.isEnabled())
            {
                synchronized(XLogBuffer.getMonitor())
                {
                    XLogBuffer.print(this.mThread.getName());
                    XLogBuffer.print(": runTimed() at ");
                    XLogBuffer.println(String.valueOf(mTargetTime));
                }
            }
        }
        mTrigger = false;
        if (CS.TIMED_THREAD_DEPENDENCIES)  ++mCalls;
        mRunnable.runTimed();
        long minTargetTime;
        if (CS.TIMED_THREAD_NANOSECONDS)
        {
            minTargetTime = nanoTime();
        }
        else
        {
            minTargetTime = System.currentTimeMillis();
        }
        if (CS.TIMED_THREAD_DILATATION)
        {
            float dilatation = sTimeDilatation;
            if (mTimeDilatationEnabled && (dilatation > 1f))
            {
                mTargetTime += (long)(dilatation * mPeriodF);
                minTargetTime += (long)(dilatation * mMinDelayF);
            }
            else
            {
                mTargetTime += mPeriodL;
                minTargetTime += mMinDelayL;
            }
        }
        else
        {
            mTargetTime += mPeriodL;
            minTargetTime += mMinDelayL;
        }
        if (mTargetTime < minTargetTime) mTargetTime = minTargetTime;
    }

    static public float getTimeDilatation()
    {
        return sTimeDilatation;
    }

    static public void setTimeDilatation(float dilatation)
    {
        sTimeDilatation = (dilatation < 1f) ? 1f : dilatation;
    }

    static private void registerThread(XTimedThread thread)
    {
        if (CS.TIMED_THREAD_DEPENDENCIES)
        {
            synchronized (sRegisteredThreads)
            {
                Iterator<XTimedThread> it_thread = sRegisteredThreads.iterator();
                while (it_thread.hasNext())
                {
                    XTimedThread thread2 = it_thread.next();
                    if (thread.mID.equals(thread2.mID)) throw new IllegalArgumentException("IDs of XTimedThread instances must be unique or null. ID: " + thread.mID);
                }
                sRegisteredThreads.add(thread);
    
                Iterator<Dependency> it_dep = sPendingDependencies.iterator();
                while (it_dep.hasNext())
                {
                    Dependency dep = it_dep.next();
                    if (dep.mThreadID.equals(thread.mID))
                    {
                        dep.mThread = thread;
                        it_dep.remove();
                    }
                }
            }
        }
    }

    static private void unregisterThread(XTimedThread thread)
    {
        if (CS.TIMED_THREAD_DEPENDENCIES)
        {
            synchronized (sRegisteredThreads)
            {
                sRegisteredThreads.remove(thread);
            }
        }
    }

    static private Dependency getDependency(String thread_id, int min_calls)
    {
        if (CS.TIMED_THREAD_DEPENDENCIES)
        {
            synchronized (sRegisteredThreads)
            {
                Iterator<XTimedThread> it = sRegisteredThreads.iterator();
                while (it.hasNext())
                {
                    XTimedThread thread = it.next();
                    if (thread.mID.equals(thread_id)) return new Dependency(thread,min_calls);
                }
    
                Dependency dep = new Dependency(thread_id,min_calls);
                sPendingDependencies.add(dep);
                return dep;
            }
        }
        else
        {
            return null;
        }
    }

    static private long nanoTime()
    {
        return System.nanoTime() - sStartNanos;
    }
}
