/********************************************************************************************
 *
 * Copyright 2012-2018  Frank Gollnik  <http://www.jabotics.org>
 *
 * 
 * This file is part of the Java software package org.jabotics (in the following: Jabotics).
 * 
 * Jabotics is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Jabotics is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jabotics.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * Diese Datei ist Teil des Java-Software-Pakets org.jabotics (im Folgenden kurz: Jabotics).
 * 
 * Jabotics ist Freie Software: Sie können es unter den Bedingungen
 * der GNU General Public License, wie von der Free Software Foundation,
 * Version 3 der Lizenz oder (nach Ihrer Wahl) jeder späteren
 * veröffentlichten Version, weiterverbreiten und/oder modifizieren.
 * 
 * Jabotics wird in der Hoffnung, dass es nützlich sein wird, aber
 * OHNE JEDE GEWÄHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 * Gewährleistung der MARKTFÄHIGKEIT oder EIGNUNG FÜR EINEN BESTIMMTEN ZWECK.
 * Siehe die GNU General Public License für weitere Details.
 * 
 * Sie sollten eine Kopie der GNU General Public License zusammen mit Jabotics
 * erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
 * 
 *******************************************************************************************/


package org.jabotics.robot.sim;

import org.jabotics.robot.defs.*;
import java.awt.geom.*;

/**
 * Tragen Sie hier eine Beschreibung des Interface ISensor ein.
 * 
 * @author  Frank Gollnik
 * @version 1.1.1
 */

abstract class BSensor extends BModelElement
{
    static private final ESensorDirection[] sDirectionsInvalidSensor = {ESensorDirection.NONE};
    static private final ESensorDirection[] sDirectionsTouchSensor = {ESensorDirection.FRONT, ESensorDirection.BACK};
    static private final ESensorDirection[] sDirectionsSoundSensor = {ESensorDirection.NONE};
    static private final ESensorDirection[] sDirectionsLightSensor = {ESensorDirection.DOWNWARD, ESensorDirection.FORWARD};
    static private final ESensorDirection[] sDirectionsUltrasonicSensor = {ESensorDirection.FORWARD, ESensorDirection.BACKWARD,ESensorDirection.TO_LEFT, ESensorDirection.TO_RIGHT};

    
    static ESensorType getDefaultSensorType(ESensorPort port)
    {
        switch (port)
        {
            case S1:
            {
                return ESensorType.TOUCH_SENSOR;
            }
            case S2:
            {
                return ESensorType.SOUND_SENSOR;
            }
            case S3:
            {
                return ESensorType.LIGHT_SENSOR;
            }
            default:
            {
                return ESensorType.ULTRASONIC_SENSOR;
            }
        }
    }

    static ESensorDirection[] getSensorDirections(ESensorType type)
    {
        switch (type)
        {
            case TOUCH_SENSOR:
            {
                return sDirectionsTouchSensor;
            }
            case SOUND_SENSOR:
            {
                return sDirectionsSoundSensor;
            }
            case LIGHT_SENSOR:
            {
                return sDirectionsLightSensor;
            }
            case ULTRASONIC_SENSOR:
            {
                return sDirectionsUltrasonicSensor;
            }
            default:
            {
                return sDirectionsInvalidSensor;
            }
        }
    }

    static ESensorDirection getDefaultSensorDirection(ESensorType type)
    {
        switch (type)
        {
            case TOUCH_SENSOR:
            {
                return ESensorDirection.BACK;
            }
            case SOUND_SENSOR:
            {
                return ESensorDirection.NONE;
            }
            case LIGHT_SENSOR:
            {
                return ESensorDirection.DOWNWARD;
            }
            case ULTRASONIC_SENSOR:
            {
                return ESensorDirection.FORWARD;
            }
            default:
            {
                return ESensorDirection.NONE;
            }
        }
    }

    static public BSensor getSensor(ESensorType type, ESensorDirection direction)
    {
        switch (type)
        {
            case TOUCH_SENSOR:
            {
                return XTouchSensor.getTouchSensor(direction);
            }
            case SOUND_SENSOR:
            {
                return XSoundSensor.getSoundSensor(direction);
            }
            case LIGHT_SENSOR:
            {
                return XLightSensor.getLightSensor(direction);
            }
            case ULTRASONIC_SENSOR:
            {
                return XUltrasonicSensor.getUltrasonicSensor(direction);
            }
            default:
            {
                return XInvalidSensor.getInvalidSensor();
            }
        }
    }

    // Instanzvariablen:
    private int mRefCount = 0;
    protected volatile XSensorRef.RefData mLeadingRefData = null;
    protected ESensorDirection mDirection = ESensorDirection.NONE;

    abstract public ESensorType getType();
    abstract public float getValue(XSensorRef ref);

    final protected void addRef() {++mRefCount;}
    final protected void releaseRef() {--mRefCount;}
    final boolean isReferenced() {return (mRefCount > 0);}
    final XSensorRef.RefData getLeadingRefData() {return mLeadingRefData;} // grahpics of sensor will reflect state leading reference
    final void setLeadingRefData(XSensorRef.RefData data) {mLeadingRefData = data;}

    final public ESensorDirection getDirection()
    {
        return mDirection;
    }

    public boolean hasLight()
    {
        return false;
    }

    public int getUpdatePeriodProlongationFactor()
    {
        return 1;
    }

    protected float getDistance(Area detectionArea, float centerX, float centerY, float[] scene_coordinates)
    {
        // determines distance from (0,0) and (if desired) scene coordinates of corresponding intersection point
        boolean determineSceneCoordinates = ( (null != scene_coordinates) && (2 == scene_coordinates.length) );
        XRobot.XDynamicData dd = XSimulationModel.getRobot().getDDSensors();
        Area area = new Area(dd.getTransformedWallPath());
        area.intersect(detectionArea);
        if (area.isEmpty()) return -1f;
        float minSqrDist = Float.MAX_VALUE;
        float sqrDist;
        float[] coordinates = new float[6];
        for (PathIterator it = area.getPathIterator(null);  !it.isDone();  it.next())
        {
            int segment_type = it.currentSegment(coordinates);
            if ( (PathIterator.SEG_MOVETO == segment_type) || (PathIterator.SEG_LINETO == segment_type) )
            {
                float dx = coordinates[0] - centerX;
                float dy = coordinates[1] - centerY;
                sqrDist = dx * dx + dy * dy;
                if (minSqrDist > sqrDist)
                {
                    minSqrDist = sqrDist;
                    if (determineSceneCoordinates)
                    {
                        // store coordinates of intersection point:
                        scene_coordinates[0] = coordinates[0]; // in robot's coordinates space
                        scene_coordinates[1] = coordinates[1];
                    }
                }
            }
        }
        if (determineSceneCoordinates)
        {
            // transform to scene coordinates:
            //???AffineTransform aft = new AffineTransform(1f, 0f, 0f, 1f, (float)dd.getRobotX(), (float)dd.getRobotY());
            AffineTransform aft = AffineTransform.getTranslateInstance(dd.getRobotX(), dd.getRobotY());
            aft.rotate(dd.getRobotPhi());
            aft.transform(scene_coordinates,0,scene_coordinates,0,1);
        }
        return (float)Math.sqrt(minSqrDist);
    }
}
