/********************************************************************************************
 *
 * Copyright 2012-2018  Frank Gollnik  <http://www.jabotics.org>
 *
 * 
 * This file is part of the Java software package org.jabotics (in the following: Jabotics).
 * 
 * Jabotics is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Jabotics is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jabotics.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * Diese Datei ist Teil des Java-Software-Pakets org.jabotics (im Folgenden kurz: Jabotics).
 * 
 * Jabotics ist Freie Software: Sie können es unter den Bedingungen
 * der GNU General Public License, wie von der Free Software Foundation,
 * Version 3 der Lizenz oder (nach Ihrer Wahl) jeder späteren
 * veröffentlichten Version, weiterverbreiten und/oder modifizieren.
 * 
 * Jabotics wird in der Hoffnung, dass es nützlich sein wird, aber
 * OHNE JEDE GEWÄHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 * Gewährleistung der MARKTFÄHIGKEIT oder EIGNUNG FÜR EINEN BESTIMMTEN ZWECK.
 * Siehe die GNU General Public License für weitere Details.
 * 
 * Sie sollten eine Kopie der GNU General Public License zusammen mit Jabotics
 * erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
 * 
 *******************************************************************************************/


package org.jabotics.robot.sim;

import org.jabotics.robot.defs.*;
import java.util.LinkedList;
import java.util.concurrent.atomic.AtomicInteger;

/**
 * Beschreiben Sie hier die Klasse XSensorRef.
 * 
 * @author  Frank Gollnik
 * @version 1.1.1
 */
class XSensorRef implements ISensor
{
    public interface IListener
    {
        void sensorRefChanged(ESensorPort port);
    }

    private static class PrefsListener implements XPreferences.IListener
    {
        private PrefsListener()
        {
        }

        @Override
        public void preferencesChanged(XPreferences.Read newPrefs)
        {
            synchronized(XSensorRef.class)
            {
                for (ESensorPort port : ESensorPort.values())
                {
                    int portID = port.ordinal();
                    RefData data = sRefData[portID];
                    BSensor oldSensor = data.mSensor;
                    ESensorType type = newPrefs.getSensorType(port);
                    ESensorDirection direction = newPrefs.getSensorDirection(port);

                    if (oldSensor.getType() != type)
                    {
                        if (!data.hasExternalRefs()) data.replaceSensor(BSensor.getSensor(type,direction));
                        
                    }
                    else if (oldSensor.getDirection() != newPrefs.getSensorDirection(port))
                    {
                        data.changeSensorDirection(direction);
                    }
                }
            }
       }
    }

    static class RefData
    {
        private final int mPortID;
        private final AtomicInteger mIntRefCount = new AtomicInteger();
        private final AtomicInteger mExtRefCount = new AtomicInteger();
        private volatile BSensor mSensor;
        private volatile boolean mLightIsOn = false;
        

        private RefData(int port_id)
        {
            mPortID = port_id;
            mSensor = null;
        }

        private RefData(int port_id, BSensor sensor)
        {
            mPortID = port_id;
            mSensor = sensor;
        }

        boolean lightIsOn()
        {
            return mLightIsOn;
        }

        private boolean hasInternalRefs()
        {
            return (mIntRefCount.get() > 0);
        }

        private void addInternalRef()
        {
            if (0 == mIntRefCount.getAndIncrement()) mSensor.addRef();
        }

        private void releaseInternalRef()
        {
            if (0 == mIntRefCount.decrementAndGet()) mSensor.releaseRef();
        }

        private boolean hasExternalRefs()
        {
            return (mExtRefCount.get() > 0);
        }

        private void addExternalRef()
        {
            if (0 == mExtRefCount.getAndIncrement())
            {
                mSensor.addRef();
                if (null == mSensor.getLeadingRefData()) mSensor.setLeadingRefData(this);
            }
        }

        private void releaseExternalRef()
        {
            if (0 == mExtRefCount.decrementAndGet())
            {
                mSensor.releaseRef();

                if (mSensor.getLeadingRefData() == this)
                {
                    mSensor.setLeadingRefData(null);
                    if (mSensor.isReferenced())
                    {
                        for (ESensorPort port : ESensorPort.values())
                        {
                            RefData data = sRefData[port.ordinal()];
                            if ((this != data) && data.hasExternalRefs() && (data.mSensor == mSensor))
                            {
                                mSensor.setLeadingRefData(data);
                                sendSensorRefChanged(port);
                                break;
                            }
                        }
                    }
                }

                sendSensorRefChanged(ESensorPort.values()[mPortID]);
            }
        }

        private void replaceSensor(BSensor newSensor)
        {
            synchronized(XSensorRef.class)
            {
                if (mSensor == newSensor) return;

                if (this.hasInternalRefs())
                {
                    mSensor.releaseRef();
                    newSensor.addRef();
                }

                if (this.hasExternalRefs())
                {
                    mSensor.releaseRef();
                    newSensor.addRef();
                }

                if (null == newSensor.getLeadingRefData()) newSensor.setLeadingRefData(this);

                if (mSensor.getLeadingRefData() == this)
                {
                    mSensor.setLeadingRefData(null);
                    if (mSensor.isReferenced())
                    {
                        for (ESensorPort port : ESensorPort.values())
                        {
                            RefData data = sRefData[port.ordinal()];
                            if ((this != data) && data.hasExternalRefs() && (data.mSensor == mSensor))
                            {
                                mSensor.setLeadingRefData(data);
                                sendSensorRefChanged(port);
                                break;
                            }
                        }
                    }
                }

                mSensor = newSensor;
                sendSensorRefChanged(ESensorPort.values()[mPortID]);
            }
        }

        private void changeSensorDirection(ESensorDirection newDirection)
        {
            synchronized(XSensorRef.class)
            {
                if (  (mSensor.getDirection() == newDirection)
                   || (mPortID < 0)                         // invalid or internal sensor reference
                   || (mPortID >= D.SENSOR_PORT_NUMBER)  )  // button reference
                   return;
        
                replaceSensor(BSensor.getSensor(mSensor.getType(),newDirection));
            }
        }
    }

    // Klassenvariablen
    private static final LinkedList<IListener> sListeners = new LinkedList<IListener>();
    private static final RefData[] sRefData = new RefData[D.TOTAL_SENSOR_NUMBER];
    private static final XSensorRef sInvalidRef = new XSensorRef();
    static
    {
        XPreferences.Read p = XPreferences.getReadInstance();
        for (ESensorPort port : ESensorPort.values())
        {
            int portID = port.ordinal();
            sRefData[portID] = new RefData(portID,BSensor.getSensor(p.getSensorType(port),p.getSensorDirection(port)));
        }
        for (EButtonType type : EButtonType.values())
        {
            int portID = type.ordinal() + D.SENSOR_PORT_NUMBER;
            sRefData[portID] = new RefData(portID,XButton.getButton(type));
        }
        XPreferences.addListener(new PrefsListener());
    }


    static XSensorRef getInvalidRef()
    {
        return sInvalidRef;
    }

    static XSensorRef getInternalRef(ESensorPort port)
    {
        return new XSensorRef(port.ordinal(),false);
    }

    static synchronized boolean hasExternalRefs(ESensorPort port)
    {
        return sRefData[port.ordinal()].hasExternalRefs();
    }

    static synchronized XSensorRef getExternalRef(ESensorPort port, ESensorType type)
    {
        if (ESensorType.INVALID_SENSOR == type)  return sInvalidRef;

        int portID = port.ordinal();
        RefData data = sRefData[portID];
        BSensor sensor = data.mSensor;
        if (data.hasExternalRefs())
        {
            if (sensor.getType() == type)
            {
                return new XSensorRef(portID,true);
            }
            else
            {
                //Zu Prüfen!
                // Should not be reached: See implementation of org.jabotics.robot.core.CSensor.getSensor(ESensorPort port, ESensorType type)
    
                // ToDo: Improve error message! Provide translation!
                System.err.println("Error: Different sensor types assigned to the same sensor port.");
                return sInvalidRef;
            }
        }
        else
        {
            if (sensor.getType() != type)
            {
                // ToDo: Improve warning! Provide translation!
                System.out.println("Warning: Assigned sensor type does not fit to sensor settings.");
                XPreferences.Write pw = XPreferences.getWriteInstance();
                ESensorDirection direction = BSensor.getDefaultSensorDirection(type);
                pw.setSensorType(port,type);
                pw.setSensorDirection(port,direction);
                pw.store();
                data.mSensor = BSensor.getSensor(type,direction);
            }
            XSensorRef ref = new XSensorRef(portID,true);
            sendSensorRefChanged(port);
            return ref;
        }
    }

    static synchronized XSensorRef getExternalRef(EButtonType type)
    {
        int portID = type.ordinal() + D.SENSOR_PORT_NUMBER;
        return new XSensorRef(portID,true);
    }



    static synchronized public void addListener(IListener listener)
    {
        if (!sListeners.contains(listener)) sListeners.add(listener);
    }

    static synchronized public boolean removeListener(IListener listener)
    {
        return sListeners.remove(listener);
    }

    static synchronized public void removeAllListeners()
    {
        sListeners.clear();
    }

    static synchronized private void sendSensorRefChanged(ESensorPort port)
    {
        for (IListener listener : sListeners)
        {
            listener.sensorRefChanged(port);
        }
    }


    // Instanzvariablen
    private volatile boolean mIsReleased = false;
    private final boolean mIsExternal;
    private final RefData mData;

    /**
     * Konstruktor für Objekte der Klasse XSensorRef
     */
    private XSensorRef()
    {
        mIsExternal = false;
        mData = new RefData(-1,BSensor.getSensor(ESensorType.INVALID_SENSOR, ESensorDirection.NONE));
    }

    private XSensorRef(int port_id, boolean is_external)
    {
        mIsExternal = is_external;
        mData = sRefData[port_id];
        if (mIsExternal)
        {
            mData.addExternalRef();
        }
        else
        {
            mData.addInternalRef();
        }
    }

    boolean sensorHasExternalRefs()
    {
        return mData.hasExternalRefs();
    }


    private void release()
    {
        if (!mIsReleased)
        {
            if (mIsExternal)
            {
                mData.releaseExternalRef();
            }
            else
            {
                mData.releaseInternalRef();
            }
            mIsReleased = true;
        }
    }

    @Override
    protected void finalize() throws Throwable
    {
        release();
        super.finalize();
    }


    boolean isInvalid()
    {
        return (mData.mSensor.getType() == ESensorType.INVALID_SENSOR);
    }

    boolean isInternal()
    {
        return (!mIsExternal && (mData.mSensor.getType() != ESensorType.INVALID_SENSOR));
    }

    boolean isExternal()
    {
        return mIsExternal;
    }

    boolean isDisplayed()
    {
        BSensor sensor = mData.mSensor;
        return  (!isInvalid() && (!mIsExternal || (sensor.getType() != ESensorType.LIGHT_SENSOR) || (mData == sensor.getLeadingRefData())));
    }

    void changeDirection(ESensorDirection newDirection)
    {
        mData.changeSensorDirection(newDirection);
    }


    @Override
    public int getPortID()
    {
        return mData.mPortID;
    }

    @Override
    public ESensorType getType()
    {
        return mData.mSensor.getType();
    }

    @Override
    public ESensorDirection getDirection()
    {
        return mData.mSensor.getDirection();
    }

    @Override
    public float getValue()
    {
        return mData.mSensor.getValue(this);
    }

    @Override
    public boolean hasLight()
    {
        return mData.mSensor.hasLight();
    }

    @Override
    public boolean lightIsOn()
    {
        return mData.mLightIsOn;
    }

    @Override
    public void lightOn()
    {
        if (mData.mSensor.hasLight()) mData.mLightIsOn = true;
    }

    @Override
    public void lightOff()
    {
        mData.mLightIsOn = false;
    }

    @Override
    public int getUpdatePeriodProlongationFactor()
    {
        return mData.mSensor.getUpdatePeriodProlongationFactor();
    }
}
