/********************************************************************************************
 *
 * Copyright 2012-2018  Frank Gollnik  <http://www.jabotics.org>
 *
 * 
 * This file is part of the Java software package org.jabotics (in the following: Jabotics).
 * 
 * Jabotics is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Jabotics is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jabotics.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * Diese Datei ist Teil des Java-Software-Pakets org.jabotics (im Folgenden kurz: Jabotics).
 * 
 * Jabotics ist Freie Software: Sie können es unter den Bedingungen
 * der GNU General Public License, wie von der Free Software Foundation,
 * Version 3 der Lizenz oder (nach Ihrer Wahl) jeder späteren
 * veröffentlichten Version, weiterverbreiten und/oder modifizieren.
 * 
 * Jabotics wird in der Hoffnung, dass es nützlich sein wird, aber
 * OHNE JEDE GEWÄHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 * Gewährleistung der MARKTFÄHIGKEIT oder EIGNUNG FÜR EINEN BESTIMMTEN ZWECK.
 * Siehe die GNU General Public License für weitere Details.
 * 
 * Sie sollten eine Kopie der GNU General Public License zusammen mit Jabotics
 * erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
 * 
 *******************************************************************************************/


package org.jabotics.robot.sim;

import org.jabotics.robot.defs.*;
import org.jabotics.robot.util.*;
import org.jabotics.robot.compile.*;

import java.util.LinkedList;
import java.util.Locale;
import java.util.concurrent.atomic.AtomicBoolean;
import java.lang.invoke.*;
import javax.swing.SwingUtilities;

/**
 * Beschreiben Sie hier die Klasse XSimulation.
 * 
 * @author  Frank Gollnik
 * @version 1.1.1
 */
public class XSimulation
{
    private XSimulation()
    {
    }

    public enum EState
    {
        INITIALIZING, PAUSING, RUNNING, COMPLETING, COMPLETED, TERMINATING
    }

    public interface IListener
    {
        void stateChanged(EState oldState, EState newState);
    }

    // constants / class variables:
    static private final String ERR_MAIN = "org.jabotics.robot.sim.XSimulation: Invalid call to main method";
    static private final String USAGE = "Usage: java -jar jabotics-sim.jar [-l <language> [-c <country>]] <robot class>";
    //static private final long UPDATEPERIOD = 10;
    //static private final long MINUPDATEPERIOD = 5;
    // XSimulation is a possible entry point into the simulation.
    // => proper initialization of simulation model needs to be secured during class initialization.
    // => can be achieved by triggering class initialization of XSession:
    static private AtomicBoolean sStarted = new AtomicBoolean(false);
    static private LinkedList<IListener> sListeners = new LinkedList<IListener>();
    static private EState sState = EState.INITIALIZING;
    static private Locale sLocale = null;
    static private XSimulationView sSimulationView = null;
    static private String sRobotClassName = null;
    static private ESimulationMode sMode = ESimulationMode.NORMAL;

    static public void main(String[] args) //throws Throwable
    {
        String language = null;
        String country = null;
        int i = 0;
        while (i < args.length)
        {
            if (args[i].equals("-h"))
            {
                if (args.length == 1)
                {
                    printUsage();
                }
                else
                {
                    printErrorMain();
                }
                return;
            }
            else if (args[i].equals("-l"))
            {
                if (i == args.length)
                {
                    printErrorMain();
                    return;
                }

                language = args[++i];
                if (language.startsWith("-"))
                {
                    printErrorMain();
                    return;
                }
            }
            else if (args[i].equals("-c"))
            {
                if (i == args.length)
                {
                    printErrorMain();
                    return;
                }

                country = args[++i];
                if (country.startsWith("-"))
                {
                    printErrorMain();
                    return;
                }
            }
            else
            {
                sRobotClassName = args[i];
                if ( (args.length - 1 != i) || sRobotClassName.startsWith("-") )
                {
                    printErrorMain();
                    return;
                }
            }
            ++i;
        }

        if ( (null == sRobotClassName)
          || ( (null == language) && (null != country) ) )
        {
            printErrorMain();
            return;
        }

        sLocale = (null == language) ? new Locale("en") : (null == country) ? new Locale(language) : new Locale(language,country);
        start();

        Class<?> robotClass;
        try
        {
            robotClass = Class.forName(sRobotClassName);
        }
        catch (ExceptionInInitializerError ie)
        {
            System.err.println(XMsg.INIT_CLASS_ERROR);
            System.err.print(XMsg.CLASS);
            System.err.print(" ");
            System.err.println(sRobotClassName);
            System.err.print(XMsg.ERROR_TYPE);
            System.err.println(" ExceptionInInitializerError");
            return;
        }
        catch (LinkageError le)
        {
            System.err.println(XMsg.INIT_CLASS_ERROR);
            System.err.print(XMsg.CLASS);
            System.err.print(" ");
            System.err.println(sRobotClassName);
            System.err.print(XMsg.ERROR_TYPE);
            System.err.println(" LinkageError");
            return;
        }
        catch (ClassNotFoundException cnfe)
        {
            System.err.println(XMsg.INIT_CLASS_ERROR);
            System.err.print(XMsg.CLASS);
            System.err.print(" ");
            System.err.println(sRobotClassName);
            System.err.print(XMsg.ERROR_TYPE);
            System.err.println(" ClassNotFoundException");
            return;
        }

        if (null == robotClass)
        {
            System.err.println(XMsg.INIT_CLASS_ERROR);
            System.err.print(XMsg.CLASS);
            System.err.print(" ");
            System.err.println(sRobotClassName);
            System.err.print(XMsg.ERROR_TYPE);
            System.err.println("Object construction failed.");
            return;
        }
        
        MethodHandles.Lookup lookup = MethodHandles.lookup();
        MethodType mt = MethodType.methodType(void.class, String[].class);
        MethodHandle mh;
        try
        {
            mh = lookup.findStatic(robotClass, "main", mt);
        }
        catch (NoSuchMethodException nsme)
        {
            System.err.println(XMsg.ACCESS_METHOD_ERROR);
            System.err.print(XMsg.CLASS);
            System.err.print(" ");
            System.err.println(sRobotClassName);
            System.err.print(XMsg.METHOD);
            System.err.println(" static main");
            System.err.print(XMsg.ERROR_TYPE);
            System.err.println(" NoSuchMethodException");
            return;
        }
        catch (IllegalAccessException iae)
        {
            System.err.println(XMsg.ACCESS_METHOD_ERROR);
            System.err.print(XMsg.CLASS);
            System.err.print(" ");
            System.err.println(sRobotClassName);
            System.err.print(XMsg.METHOD);
            System.err.println(" static main");
            System.err.print(XMsg.ERROR_TYPE);
            System.err.println(" IllegalAccessException");
            return;
        }
        if (null == mh)
        {
            System.err.println(XMsg.ACCESS_METHOD_ERROR);
            System.err.print(XMsg.CLASS);
            System.err.print(" ");
            System.err.println(sRobotClassName);
            System.err.print(XMsg.METHOD);
            System.err.println(" static main");
            System.err.print(XMsg.ERROR_TYPE);
            System.err.println(" Creation of method handle failed.");
            return;
        }

        do
        {
            try
            {
                mh.invoke((Object[])null);
            }
            catch (Throwable t)
            {
                System.err.println(XMsg.ACCESS_METHOD_ERROR);
                System.err.print(XMsg.CLASS);
                System.err.print(" ");
                System.err.println(sRobotClassName);
                System.err.print(XMsg.METHOD);
                System.err.println(" static main");
                System.err.print(XMsg.ERROR_TYPE);
                System.err.print(" ");
                System.err.println(t.getMessage());
                return;
                //throw (t);
            }

            if (CS.LOG_BUFFER)
            {
                XLogBuffer.enableLogging(false);
                System.out.println(XLogBuffer.flush());
            }

            completeSimulation();
        }
        while (ESimulationMode.NORMAL == sMode); 
    }

    static boolean mainIsRunning()
    {
        return (null != sRobotClassName);
    }

    static private void printErrorMain()
    {
        System.err.println(ERR_MAIN);
        System.err.println(USAGE);
    }

    static private void printUsage()
    {
        System.out.println(USAGE);
    }

    static void start()
    {
        if (sStarted.compareAndSet(false,true))
        {
            if (!mainIsRunning() && (ESimulationMode.HIDDEN != sMode))
            {
                // Simulation started by robot class (possibly without a call to XSimulation.main):
                // Requires joined thread to complete simulation after termination of robot thread
                (new Thread("Complete simulation thread")
                {
                    private Thread mThread = Thread.currentThread();
                    public void run()
                    {
                        try
                        {
                            mThread.join();
                            // Check once again if simulation has been started implicitly without a call to XSimulation.main:
                            if (!mainIsRunning()) completeSimulation();
                        }
                        catch (Exception e) {}
                    }
                }).start();
            }
            if (CS.LOG_BUFFER)
            {
                XLogBuffer.enableLogging(true);
            }
            if (null == sLocale) sLocale = Locale.getDefault();
            XMsg.init(sLocale);
            XTimer.setRobotThread(Thread.currentThread());
            if (ESimulationMode.HIDDEN != sMode) showGraphics(ESimulationMode.NORMAL == sMode);
            if (ESimulationMode.NORMAL == sMode)
            {
                XTimer.msDelayRobot(1);  // This will make sure that simulation is initially pausing
            }
            else
            {
                resumeSimulation();
            }
        }
    }

    static private void completeSimulation()
    {
        if (ESimulationMode.HIDDEN != sMode) // hidden mode will be possible here in special scenarios (like test class
                                             // setting hidden mode first and then making a call to XSimulation.main)
        {
            setState(EState.COMPLETING);
            XTimer.pauseRobot(false);
            XTimer.msDelay(500);
            XAudioSystem.playRobotSignal();
            setState(EState.COMPLETED);
    
            if (ESimulationMode.NORMAL == sMode)
            {
                while (EState.INITIALIZING != sState) XTimer.msDelay(10);
        
                if (CS.LOG_BUFFER)
                {
                    XLogBuffer.enableLogging(true);
                }

                XTimer.msDelayRobot(1);  // This will make sure that simulation is initially pausing
            }
        }
    }

    static private void showGraphics(final boolean interactive)
    {
        if (null == sSimulationView)
        {
            SwingUtilities.invokeLater(new Runnable()
            {
                public void run()
                {
                    sSimulationView = new XSimulationView("Jabotics - " + XMsg.ROBOT_SIM + " - " + sRobotClassName, interactive);
                    sSimulationView.setVisible(true);
                    sSimulationView.startUpdate();
                }
            });
        }
    }

    static private void setState(EState newState)
    {
        if (newState != sState)
        {
            EState oldState = sState;
            sState = newState;
            for (IListener listener : sListeners)
            {
                listener.stateChanged(oldState,newState);
            }
        }
    }


    static synchronized void addListener(IListener listener)
    {
        if (!sListeners.contains(listener)) sListeners.add(listener);
    }

    static synchronized boolean removeListener(IListener listener)
    {
        return sListeners.remove(listener);
    }

    static synchronized void removeAllListeners()
    {
        sListeners.clear();
    }

    static Locale getLocale()
    {
        return sLocale;
    }

    // This method will be called by the UI to stop the simulation process.
    static synchronized void terminateSimulationProcess()
    {
        XTimer.resumeRobotThread();
        setState(EState.TERMINATING);
        System.exit(0);
    }

    static EState getState()
    {
        return sState;
    }

    static void resetRobot()
    {
        if (EState.COMPLETED == sState)
        {
            XSession.getSession().reset();
            setState(EState.INITIALIZING);
        }
    }

    static boolean pauseSimulation()
    {
        if (EState.RUNNING == sState)
        {
            XTimer.pauseRobot(XPreferences.getReadInstance().getSimSuspendEnabled());
            setState(EState.PAUSING);
            return true;
        }
        
        return false;
    }

    static boolean resumeSimulation()
    {
        if ((EState.PAUSING == sState) || (EState.INITIALIZING == sState))
        {
            XTimer.resumeRobot();
            setState(EState.RUNNING);
            return true;
        }
        
        return false;
    }

    static void setMode(ESimulationMode mode)
    {
        if (!sStarted.get()) sMode = mode;
    }

    static public void test()
    {
        sLocale = new Locale("en","US");
        XMsg.init(sLocale);
        start();
        //resumeSimulation();
        ISensor sensor1 = XSensorRef.getExternalRef(ESensorPort.S1,ESensorType.TOUCH_SENSOR);
        ISensor sensor2 = XSensorRef.getExternalRef(ESensorPort.S2,ESensorType.SOUND_SENSOR);
        ISensor sensor3 = XSensorRef.getExternalRef(ESensorPort.S3,ESensorType.LIGHT_SENSOR);
        ISensor sensor4 = XSensorRef.getExternalRef(ESensorPort.S4,ESensorType.ULTRASONIC_SENSOR);
        float value1 = sensor1.getValue();
        float value2 = sensor2.getValue();
        float value3 = sensor3.getValue();
        float value4 = sensor4.getValue();
        System.out.println("Sensor 1: value = " + value1);
        System.out.println("Sensor 2: value = " + value2);
        System.out.println("Sensor 3: value = " + value3);
        System.out.println("Sensor 4: value = " + value4);
    }
}
