/********************************************************************************************
 *
 * Copyright 2012-2018  Frank Gollnik  <http://www.jabotics.org>
 *
 * 
 * This file is part of the Java software package org.jabotics (in the following: Jabotics).
 * 
 * Jabotics is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Jabotics is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jabotics.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * Diese Datei ist Teil des Java-Software-Pakets org.jabotics (im Folgenden kurz: Jabotics).
 * 
 * Jabotics ist Freie Software: Sie können es unter den Bedingungen
 * der GNU General Public License, wie von der Free Software Foundation,
 * Version 3 der Lizenz oder (nach Ihrer Wahl) jeder späteren
 * veröffentlichten Version, weiterverbreiten und/oder modifizieren.
 * 
 * Jabotics wird in der Hoffnung, dass es nützlich sein wird, aber
 * OHNE JEDE GEWÄHRLEISTUNG, bereitgestellt; sogar ohne die implizite
 * Gewährleistung der MARKTFÄHIGKEIT oder EIGNUNG FÜR EINEN BESTIMMTEN ZWECK.
 * Siehe die GNU General Public License für weitere Details.
 * 
 * Sie sollten eine Kopie der GNU General Public License zusammen mit Jabotics
 * erhalten haben. Wenn nicht, siehe <http://www.gnu.org/licenses/>.
 * 
 *******************************************************************************************/


package org.jabotics.robot.sim;

import org.jabotics.robot.defs.ISessionData;
import org.jabotics.robot.util.*;

import java.util.ArrayList;
import java.util.Iterator;
//import javax.sound.sampled.*;

/**
 * Beschreiben Sie hier die Klasse XSoundscape.
 * 
 * @author  Frank Gollnik
 * @version 1.1.1
 */
class XSoundscape implements XTimedThread.Runnable, XPreferences.IListener, ISessionData
{
    // constants and class variables:
    static private final int MIXER_THREAD_PRIORITY = Thread.MAX_PRIORITY - 3;
    static private final int UPDATE_DELAY = 1;
    static private final float NOISE_AMPLITUDE = 10f;
    static private final float NOISE_PERIOD = 0.05f; // in s
    static private final float CLAP_PERIOD = 0.15f;  // in s

    private class SoundInfo
    {
        ArrayList<Float> mLevels = new ArrayList<Float>();
        int mTotalLevel = -1;
    }

    private class Task
    {
        volatile Task mPrev = null;
        volatile Task mNext = null;
        boolean mClearNoise = false;
        SoundSource mSoundSource = null;
    }

    private class SoundSource
    {
        private float mLevel;
        private float mAmplitude;
        private double mPeriod; // in s
        private double mTargetTime;
        private float mTargetLevel;
        private float mCurrentLevel;

        SoundSource(float level, double period) // generates a clap
        {
            mLevel = 0f;
            mAmplitude = 0f;
            mPeriod = period;
            mTargetTime = 0.;
            mTargetLevel = level;
            mCurrentLevel = 0f;
        }

        SoundSource(float level, float amplitude, double period) // generates noise
        {
            mLevel = level;
            mAmplitude = amplitude;
            mPeriod = period;
            mTargetTime = 0.;
            mTargetLevel = level;
            mCurrentLevel = level;
        }

        boolean update()
        {
            if (mTargetTime < XSimulationModel.sTime)
            {
                if (mLevel <= 0)
                {
                    // clap
                    if (mTargetLevel <= 0) return true; // second phase is over => done

                    // start second phase (decay)
                    mTargetLevel = 0f;
                    mTargetTime = 10. * mPeriod;
                }
                else
                {
                    // noise
                    mTargetLevel = mLevel + (float)(2.*Math.random() - 1.) * mAmplitude;
                    mTargetTime = XSimulationModel.sTime + mPeriod;
                }
            }
            else
            {
                float factorReplacing = (float)(XSimulationModel.sDT / mPeriod);
                if (factorReplacing > 1f) factorReplacing = 1f;
                float factorRemaining = 1f - factorReplacing;
                mCurrentLevel = factorRemaining * mCurrentLevel + factorReplacing * mTargetLevel;
            }
            return false;
        }
    }

    static private XSoundscape sSoundscape;

    private XTimedThread mThread;
    private volatile ArrayList<SoundSource> mSources;
    private volatile SoundInfo mSoundInfo;
    private volatile int mInputLevel;
    private volatile Task mLastTask;
    private volatile Task mFirstTask;
    private volatile int mMicSignal;
    private volatile boolean mStopController = false;
    private volatile boolean mControllerHalted = false;

    static synchronized XSoundscape getSoundscape()
    {
        if (null == sSoundscape)  sSoundscape = new XSoundscape();
        return sSoundscape;
    }

    private XSoundscape()
    {
        initVolatileMembers();
        mThread = new XTimedThread("XSoundscape controler thread",this);
        mThread.enableTimeDilatation(false);
        mThread.setDaemon(true);
        mThread.setPriority(MIXER_THREAD_PRIORITY);
        mThread.setTimeouts(UPDATE_DELAY,UPDATE_DELAY,0);
        mThread.start();
        XPreferences.addListener(this);
        XSession.getSession().add(this);
    }
    
    private void initVolatileMembers()
    {
        XPreferences.Read p = XPreferences.getReadInstance();
        mSources = new ArrayList<SoundSource>();
        mSoundInfo = new SoundInfo();
        mInputLevel = p.getAudioInputLevel();
        mLastTask = new Task();
        mFirstTask = mLastTask;
        mMicSignal = 0;
        this.addNoise(mInputLevel, (mInputLevel < NOISE_AMPLITUDE) ? mInputLevel : NOISE_AMPLITUDE, NOISE_PERIOD);
    }

    /*private void clearNoise()
    {
        Task task = new Task();
        task.mClearNoise = true;
        synchronized (this)
        {
            task.mPrev = mLastTask;
            mLastTask.mNext = task;
            mLastTask = task;
        }
    }*/

    private void addNoise(float level, float amplitude, double period)
    {
        Task task = new Task();
        task.mSoundSource = new SoundSource(level, amplitude, period);
        synchronized (this)
        {
            task.mPrev = mLastTask;
            mLastTask.mNext = task;
            mLastTask = task;
        }
    }

    private void replaceNoise(float level, float amplitude, double period)
    {
        Task task = new Task();
        task.mClearNoise = true;
        task.mSoundSource = new SoundSource(level, amplitude, period);
        synchronized (this)
        {
            task.mPrev = mLastTask;
            mLastTask.mNext = task;
            mLastTask = task;
        }
    }

    void addClap(int level)
    {
        if (level <= 0) return;
        addClap(1.6f*level,CLAP_PERIOD); // empirical factor to reach passed value as peak value ( 1/(1-1/e) = 1.58 )
    }

    private void addClap(float level, double period)
    {
        Task task = new Task();
        task.mSoundSource = new SoundSource(level, period);
        synchronized (this)
        {
            task.mPrev = mLastTask;
            mLastTask.mNext = task;
            mLastTask = task;
        }
    }

    @Override
    public synchronized void init()
    {
        this.runTimed(); // single call to complete initialization
    }

    @Override
    public synchronized void shutdown()
    {
    }

    @Override
    public synchronized void terminate()
    {
    }

    @Override
    public synchronized void runTimed()
    {
        mControllerHalted = mStopController;
        if (mStopController) return;

        mMicSignal = XAudioSystem.getMicSignal();

        // process list of new tasks in any case (preferences could change: keep track of sound input level respectively noise level)
        while (mFirstTask != mLastTask)
        {
            Task first = mFirstTask;
            mFirstTask = first.mNext;
            mFirstTask.mPrev = null;
            first.mNext = null;

            if (mFirstTask.mClearNoise)
            {
                for (Iterator<SoundSource> it = mSources.iterator(); it.hasNext(); )
                {
                    if (it.next().mLevel > 0)  it.remove();
                }
            }

            SoundSource soundSource = mFirstTask.mSoundSource;
            if (null != soundSource) 
            {
                soundSource.mTargetTime = XSimulationModel.sTime + soundSource.mPeriod;
                mSources.add(soundSource);
            }
        }
    
        if (mMicSignal < 0)
        {
            // mic is off  =>  simulate noise / clap
            SoundInfo info = new SoundInfo();
            for (Iterator<SoundSource> it = mSources.iterator(); it.hasNext(); )
            {
                SoundSource source = it.next();
                if (source.update())
                {
                    // completed
                    it.remove();
                }
                else
                {
                    info.mLevels.add(source.mCurrentLevel);
                }
            }
            mSoundInfo = info;
        }
    }

    int getSoundLevel()
    {
        int micSignal = mMicSignal;

        if (micSignal < 0)
        {
            SoundInfo info = mSoundInfo;
            synchronized(info)
            {
                if (info.mTotalLevel < 0)
                {
                    double totalLevel = 0.;
                    for (float level : info.mLevels)
                    {
                        if (level > 0) totalLevel += Math.exp(level);
                    }
                    info.mTotalLevel = (totalLevel > 0) ? (int)Math.log(totalLevel) : 0;
                    if (info.mTotalLevel < 0)
                    {
                        info.mTotalLevel = 0;
                    }
                    else if (info.mTotalLevel > 100)
                    {
                        info.mTotalLevel = 100;
                    }
                }

                return info.mTotalLevel;
            }
        }
        else
        {
            int level = (micSignal * mInputLevel) / 50;  // => max "amplification" by factor of 100/50 = 2
            return (level < 0) ? 0 : ( (level > 100) ? 100 : level );
        }
    }

    public void preferencesChanged(XPreferences.Read newPrefs)
    {
        if (mInputLevel != newPrefs.getAudioInputLevel())
        {
            mInputLevel = newPrefs.getAudioInputLevel();
            this.replaceNoise(mInputLevel, (mInputLevel < NOISE_AMPLITUDE) ? mInputLevel : NOISE_AMPLITUDE, NOISE_PERIOD);
            if (XSimulation.EState.RUNNING != XSimulation.getState())
            {
                runTimed(); // manual call necessary in this case to complete initialization
            }
        }
    }

    @Override
    public String getKey()
    {
        return "XSoundscape";
    }

    @Override
    public void stopSession()
    {
        try
        {
            mStopController = true;
            while (!mControllerHalted) Thread.sleep(1);
        }
        catch (Exception e) {}
    }

    @Override
    public ISessionData resetSession()
    {
        initVolatileMembers();
        return this;
    }

    @Override
    public void startSession()
    {
        mStopController = false;
        mControllerHalted = false; // do not wait for controller thread to update sControllerHalted
    }

    static public void test1()
    {
        XSoundscape soundscape = XSoundscape.getSoundscape();
        soundscape.runTimed();
        int level = soundscape.getSoundLevel();
        System.out.println("Sound level = " + level);
    }
}
